-------------------------------------------------------------------------------
-- Title      : Mock Turtle Core
-- Project    : Mock Turtle
-------------------------------------------------------------------------------
-- File       : mock_turtle_core.vhd
-- Author     : Tomasz Włostowski
-- Company    : CERN BE-CO-HT
-- Created    : 2014-04-01
-- Last update: 2018-03-19
-- Platform   : FPGA-generic
-- Standard   : VHDL'93
-------------------------------------------------------------------------------
-- Description:
--
-- Mock Turtle Core - top level, interconnecting the CPU cores,
-- Message Queues, Host interface and the Shared Memory.
-------------------------------------------------------------------------------
--
-- Copyright (c) 2014 CERN
--
-- This source file is free software; you can redistribute it
-- and/or modify it under the terms of the GNU Lesser General
-- Public License as published by the Free Software Foundation;
-- either version 2.1 of the License, or (at your option) any
-- later version.
--
-- This source is distributed in the hope that it will be
-- useful, but WITHOUT ANY WARRANTY; without even the implied
-- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
-- PURPOSE.  See the GNU Lesser General Public License for more
-- details.
--
-- You should have received a copy of the GNU Lesser General
-- Public License along with this source; if not, download it
-- from http://www.gnu.org/licenses/lgpl-2.1.html
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.mock_turtle_pkg.all;
use work.wishbone_pkg.all;
use work.mt_cpu_csr_wbgen2_pkg.all;
use work.mt_per_cpu_csr_pkg.all;
use work.mt_mqueue_pkg.all;
use work.mt_private_pkg.all;
use work.gencores_pkg.all;
use work.wr_fabric_pkg.all;

entity mock_turtle_core is

  generic (
-- Message Queue and CPU configuration
    g_config            : t_mock_turtle_config := c_default_mock_turtle_config;
-- When true, the Remote Message Queue is implemented.
    g_with_rmq          : boolean          := true;
    g_use_wr_fabric     : boolean          := true;
-- Frequency of clk_sys_i, in Hz
    g_system_clock_freq : integer          := 62500000;
-- Enables/disables WR support
    g_with_white_rabbit : boolean          := false);

  port (
    clk_i     : in std_logic;
    rst_n_i   : in std_logic;

    -- shared peripheral port
    sp_master_o : out t_wishbone_master_out;
    sp_master_i : in  t_wishbone_master_in := cc_dummy_master_in;

    -- dedicated (per-cpu) peripheral port
    dp_master_o : out t_wishbone_master_out_array(0 to g_config.cpu_count-1);
    dp_master_i : in  t_wishbone_master_in_array(0 to g_config.cpu_count-1) := f_dummy_master_in_array(g_config.cpu_count);

    rmq_src_o : out t_mt_stream_source_out;
    rmq_src_i : in  t_mt_stream_source_in := c_mt_dummy_source_in;
    rmq_snk_o : out t_mt_stream_sink_out;
    rmq_snk_i : in  t_mt_stream_sink_in   := c_mt_dummy_sink_in;

    wr_src_o : out t_wrf_source_out;
    wr_src_i : in  t_wrf_source_in := c_dummy_src_in;
    wr_snk_o : out t_wrf_sink_out;
    wr_snk_i : in  t_wrf_sink_in   := c_dummy_snk_in;

    host_slave_i : in  t_wishbone_slave_in;
    host_slave_o : out t_wishbone_slave_out;

    clk_ref_i : in std_logic := '0';
    tm_i      : in t_mt_timing_if;

    gpio_o : out std_logic_vector(31 downto 0);
    gpio_i : in  std_logic_vector(31 downto 0);

    rmq_swrst_o : out std_logic;

    host_irq_o      : out std_logic;
    debug_msg_irq_o : out std_logic
    );

end mock_turtle_core;

architecture rtl of mock_turtle_core is

  ------------------------------------------
  --        COMPONENTs DECLARATION
  ------------------------------------------

  component mt_mqueue_host
    generic (
      g_config : t_mt_mqueue_config);
    port (
      clk_i        : in  std_logic;
      rst_n_i      : in  std_logic;
      si_slave_i   : in  t_wishbone_slave_in;
      si_slave_o   : out t_wishbone_slave_out;
      host_slave_i : in  t_wishbone_slave_in;
      host_slave_o : out t_wishbone_slave_out;
      host_irq_o   : out std_logic;
      hmq_status_o : out std_logic_vector(15 downto 0));
  end component;

  component mt_mqueue_remote is
    generic (
      g_config        : t_mt_mqueue_config;
      g_use_wr_fabric : boolean);
    port (
      clk_i        : in  std_logic;
      rst_n_i      : in  std_logic;
      si_slave_i   : in  t_wishbone_slave_in;
      si_slave_o   : out t_wishbone_slave_out;
      src_o        : out t_mt_stream_source_out;
      src_i        : in  t_mt_stream_source_in;
      snk_o        : out t_mt_stream_sink_out;
      snk_i        : in  t_mt_stream_sink_in;
      wr_snk_i     : in  t_wrf_sink_in   := c_dummy_snk_in;
      wr_snk_o     : out t_wrf_sink_out;
      wr_src_i     : in  t_wrf_source_in := c_dummy_src_in;
      wr_src_o     : out t_wrf_source_out;
      rmq_swrst_o  : out std_logic;
      rmq_status_o : out std_logic_vector(15 downto 0);
      debug_o      : out std_logic_vector(31 downto 0));
  end component mt_mqueue_remote;

  component mt_trace_profiler is
    generic (
      g_config : t_mock_turtle_config);
    port (
      clk_i          : in  std_logic;
      rst_n_i        : in  std_logic;
      cpu_pc_i       : in  t_pc_array(0 to g_config.cpu_count-1);
      cpu_pc_valid_i : in  std_logic_vector(g_config.cpu_count-1 downto 0);
      slave_i        : in  t_wishbone_slave_in;
      slave_o        : out t_wishbone_slave_out);
  end component mt_trace_profiler;

  component mt_shared_mem is
    generic (
      g_size : integer);
    port (
      clk_i   : in  std_logic;
      rst_n_i : in  std_logic;
      slave_i : in  t_wishbone_slave_in;
      slave_o : out t_wishbone_slave_out);
  end component mt_shared_mem;

  component mt_wb_remapper is
    generic (
      g_num_ranges : integer;
      g_base_in    : t_wishbone_address_array;
      g_base_out   : t_wishbone_address_array;
      g_mask_in    : t_wishbone_address_array;
      g_mask_out   : t_wishbone_address_array);
    port (
      slave_i  : in  t_wishbone_slave_in;
      slave_o  : out t_wishbone_slave_out;
      master_i : in  t_wishbone_master_in;
      master_o : out t_wishbone_master_out);
  end component mt_wb_remapper;

  ------------------------------------------
  --        CONSTANTS DECLARATION
  ------------------------------------------

  constant c_smem_remap_mask_in : t_wishbone_address_array(4 downto 0) := (
    0 => x"0001c000",
    1 => x"0001c000",
    2 => x"0001c000",
    3 => x"0001c000",
    4 => x"00010000"
    );
  constant c_smem_remap_mask_out : t_wishbone_address_array(4 downto 0) := (
    0 => x"00003fff",
    1 => x"00003fff",
    2 => x"00003fff",
    3 => x"00003fff",
    4 => x"0000ffff"
    );

  constant c_smem_remap_base_in : t_wishbone_address_array(4 downto 0) := (
    0 => x"00000000",
    1 => x"00004000",
    2 => x"00008000",
    3 => x"0000c000",
    4 => x"00010000"
    );

  -- remapping to squeeze the host address space to 128 kB
  constant c_smem_remap_base_out : t_wishbone_address_array(4 downto 0) := (
    0 => x"00000000",                   -- 0x0000-0x3fff -> HMQ GCR
    1 => x"00004000",                   -- 0x4000-0x7fff -> HMQ IN (muxed)
    2 => x"00008000",                   -- 0x8000-0xbfff -> HMQ OUT (muxed)
    3 => x"00010000",                   -- 0xc000-0xffff -> other host peripherals
    4 => x"00200000"                    -- 0x10000-0x1ffff -> SMEM
    );


  constant c_hac_wishbone_masters : integer := 5;
  constant c_hac_master_hmq       : integer := 0;
  constant c_hac_master_cpu_csr   : integer := 1;
  constant c_hac_master_tpu       : integer := 2;
  constant c_hac_master_crom      : integer := 3;
  constant c_hac_master_si        : integer := 4;


  constant c_hac_address : t_wishbone_address_array (c_hac_wishbone_masters-1 downto 0) := (
    c_hac_master_hmq     => x"00000000",  -- Host MQ
    c_hac_master_cpu_csr => x"00010000",  -- CPU CSR
    c_hac_master_tpu     => x"00011000",  -- Trace/Profile Unit
    c_hac_master_crom    => x"00012000",  -- Config ROM
    c_hac_master_si      => x"00200000"   -- SMEM
    );

  constant c_hac_mask : t_wishbone_address_array (c_hac_wishbone_masters-1 downto 0) := (
    c_hac_master_hmq     => x"003f0000",  -- Host MQ
    c_hac_master_cpu_csr => x"003ff000",  -- CPU CSR
    c_hac_master_tpu     => x"003ff000",  -- Trace/Profile Unit
    c_hac_master_crom    => x"003ffc00",  -- Config ROM
    c_hac_master_si      => x"00300000"   -- SMEM
    );

  signal hac_master_out : t_wishbone_master_out_array(c_hac_wishbone_masters-1 downto 0);
  signal hac_master_in  : t_wishbone_master_in_array(c_hac_wishbone_masters-1 downto 0);

  constant c_si_wishbone_masters : integer := 5;
  constant c_si_wishbone_slaves  : integer := g_config.cpu_count + 1;
  constant c_si_master_hmq       : integer := 0;
  constant c_si_master_rmq       : integer := 1;
  constant c_si_master_smem      : integer := 2;
  constant c_si_master_crom      : integer := 3;
  constant c_si_master_sp        : integer := 4;

  constant c_si_slave_hac  : integer := 0;
  constant c_si_slave_cpu0 : integer := 1;

  constant c_si_address : t_wishbone_address_array (c_si_wishbone_masters-1 downto 0) := (
    c_si_master_hmq  => x"00000000",    -- Host MQ
    c_si_master_rmq  => x"00100000",    -- Remote MQ
    c_si_master_smem => x"00200000",    -- Shared Memory
    c_si_master_crom => x"00400000",    -- Config ROM
    c_si_master_sp   => x"08000000"
    );

  constant c_si_mask : t_wishbone_address_array (c_si_wishbone_masters-1 downto 0) := (
    c_si_master_hmq  => x"0ff00000",    -- Host MQ
    c_si_master_rmq  => x"0ff00000",    -- Remote MQ
    c_si_master_smem => x"0ff00000",    -- Shared Memory
    c_si_master_crom => x"0ffffc00",    -- Config ROM
    c_si_master_sp   => x"08000000"
    );

  signal si_slave_in  : t_wishbone_slave_in_array(c_si_wishbone_slaves-1 downto 0);
  signal si_slave_out : t_wishbone_slave_out_array(c_si_wishbone_slaves-1 downto 0);

  signal si_master_in  : t_wishbone_master_in_array(c_si_wishbone_masters-1 downto 0);
  signal si_master_out : t_wishbone_master_out_array(c_si_wishbone_masters-1 downto 0);

  subtype t_cpu_rng is natural range g_config.cpu_count-1 downto 0;
  subtype t_max_cpu_rng is natural range 7 downto 0;

  type t_word_array is array(integer range <>) of std_logic_vector(31 downto 0);
  type t_byte_array is array(integer range <>) of std_logic_vector(7 downto 0);

  --  CSR inputs/outputs
  signal cpu_csr_fromwb : t_mt_cpu_csr_out_registers;
  signal cpu_csr_towb   : t_mt_cpu_csr_in_registers;

  type t_mt_per_cpu_csr_in_array is array(t_cpu_rng) of t_mt_per_cpu_csr_in;
  type t_mt_per_cpu_csr_out_array is array(t_cpu_rng) of t_mt_per_cpu_csr_out;

  signal per_cpu_csr_towb   : t_mt_per_cpu_csr_in_array;
  signal per_cpu_csr_fromwb : t_mt_per_cpu_csr_out_array;

  signal mbx_data_fromwb : t_word_array(t_max_cpu_rng);
  signal mbx_wr : std_logic_vector(t_max_cpu_rng);
  signal mbx_data_towb : t_word_array(t_max_cpu_rng);
  signal dbg_insn : t_word_array(t_max_cpu_rng);
  signal dbg_insn_wr : std_logic_vector(t_max_cpu_rng);
  signal dbg_insn_ready : std_logic_vector(t_max_cpu_rng);

  signal hmq_status, rmq_status : std_logic_vector(15 downto 0);

  signal cpu_index : integer := 0;

  signal cpu_dbg_drdy, cpu_dbg_dack : std_logic_vector(t_cpu_rng);
  signal cpu_dbg_msg_data           : t_byte_array(t_cpu_rng);
  signal dbg_msg_data_read_ack      : std_logic;

  signal cpu_gpio_out    : t_word_array (t_cpu_rng);

  signal rst_n_ref : std_logic;

  signal host_remapped_in  : t_wishbone_slave_in;
  signal host_remapped_out : t_wishbone_slave_out;

  signal cpu_pcs_valid : std_logic_vector(t_cpu_rng);
  signal cpu_pcs       : t_pc_array(0 to g_config.cpu_count-1);

  ------------------------------------------
  --        FUNCTIONS DECLARATION
  ------------------------------------------
  function f_reduce_or (x : t_word_array) return std_logic_vector is
    variable rv : std_logic_vector(31 downto 0);
  begin
    rv := (others => '0');
    for n in x'range loop
      rv := rv or x(n);
    end loop;
    return rv;
  end f_reduce_or;

begin  -- rtl

  U_Sync_Refclk : gc_sync_ffs
    port map (
      clk_i    => clk_ref_i,
      rst_n_i  => '1',
      data_i   => rst_n_i,
      synced_o => rst_n_ref);


  U_Remap_SMEM : mt_wb_remapper
    generic map (
      g_num_ranges => 5,
      g_base_in    => c_smem_remap_base_in,
      g_base_out   => c_smem_remap_base_out,
      g_mask_in    => c_smem_remap_mask_in,
      g_mask_out   => c_smem_remap_mask_out)
    port map (
      slave_i  => host_slave_i,
      slave_o  => host_slave_o,
      master_i => host_remapped_out,
      master_o => host_remapped_in);

  U_Host_Access_CB : xwb_crossbar
    generic map (
      g_num_masters => 1,
      g_num_slaves  => c_hac_wishbone_masters,
      g_registered  => true,
      g_address     => c_hac_address,
      g_mask        => c_hac_mask)
    port map (
      clk_sys_i  => clk_i,
      rst_n_i    => rst_n_i,
      slave_i(0) => host_remapped_in,
      slave_o(0) => host_remapped_out,
      master_i   => hac_master_in,
      master_o   => hac_master_out);


  U_Shared_Interconnect : xwb_crossbar
    generic map (
      g_num_masters => c_si_wishbone_slaves,
      g_num_slaves  => c_si_wishbone_masters,
      g_registered  => true,
      g_address     => c_si_address,
      g_mask        => c_si_mask)
    port map (
      clk_sys_i => clk_i,
      rst_n_i   => rst_n_i,
      slave_i   => si_slave_in,
      slave_o   => si_slave_out,
      master_i  => si_master_in,
      master_o  => si_master_out);

  gen_with_tpu : if g_config.tpu_enable generate

    U_TPU : mt_trace_profiler
      generic map (
        g_config => g_config)
      port map (
        clk_i          => clk_i,
        rst_n_i        => rst_n_i,
        cpu_pc_i       => cpu_pcs,
        cpu_pc_valid_i => cpu_pcs_valid,
        slave_i => hac_master_out(c_hac_master_tpu),
        slave_o => hac_master_in(c_hac_master_tpu)
        );

  end generate gen_with_tpu;

  gen_without_tpu : if not g_config.tpu_enable generate
    hac_master_in(c_hac_master_tpu) <= cc_dummy_master_in;
  end generate gen_without_tpu;

  sp_master_o                   <= si_master_out(c_si_master_sp);
  si_master_in (c_si_master_sp) <= sp_master_i;

-- hack: replace SMEM high address bits with the SMEM_OP register value. This way,
-- the entire 64 kB SMEM window can be visible to the host with direct addressing,
-- and only the type of atomic operation has to be chosen indirectly (by
-- writing to SMEM_OP).

  si_slave_in(c_si_slave_hac).cyc               <= hac_master_out(c_hac_master_si).cyc;
  si_slave_in(c_si_slave_hac).stb               <= hac_master_out(c_hac_master_si).stb;
  si_slave_in(c_si_slave_hac).we                <= hac_master_out(c_hac_master_si).we;
  si_slave_in(c_si_slave_hac).sel               <= hac_master_out(c_hac_master_si).sel;
  si_slave_in(c_si_slave_hac).dat               <= hac_master_out(c_hac_master_si).dat;
  si_slave_in(c_si_slave_hac).adr(15 downto 0)  <= hac_master_out(c_hac_master_si).adr(15 downto 0);
  si_slave_in(c_si_slave_hac).adr(19 downto 16) <= '0' & cpu_csr_fromwb.smem_op_o;
  si_slave_in(c_si_slave_hac).adr(31 downto 20) <= x"002";

  hac_master_in(c_hac_master_si) <= si_slave_out(c_si_slave_hac);


  U_CPU_CSR : entity work.mt_cpu_csr_wb_slave
    port map (
      rst_n_i               => rst_n_i,
      clk_sys_i             => clk_i,
      wb_adr_i              => hac_master_out(c_hac_master_cpu_csr).adr(6 downto 2),
      wb_dat_i              => hac_master_out(c_hac_master_cpu_csr).dat,
      wb_dat_o              => hac_master_in(c_hac_master_cpu_csr).dat,
      wb_cyc_i              => hac_master_out(c_hac_master_cpu_csr).cyc,
      wb_sel_i              => hac_master_out(c_hac_master_cpu_csr).sel,
      wb_stb_i              => hac_master_out(c_hac_master_cpu_csr).stb,
      wb_we_i               => hac_master_out(c_hac_master_cpu_csr).we,
      wb_ack_o              => hac_master_in(c_hac_master_cpu_csr).ack,
      wb_stall_o            => hac_master_in(c_hac_master_cpu_csr).stall,
      dbg_msg_data_rd_ack_o => dbg_msg_data_read_ack,
      regs_i                => cpu_csr_towb,
      regs_o                => cpu_csr_fromwb);

  hac_master_in(c_hac_master_cpu_csr).err <= '0';
  hac_master_in(c_hac_master_cpu_csr).rty <= '0';

  cpu_index <= to_integer(unsigned(cpu_csr_fromwb.core_sel_o));

  cpu_csr_towb.app_id_i       <= g_config.app_id;
  cpu_csr_towb.core_count_i   <= std_logic_vector(to_unsigned(g_config.cpu_count, 4));
  cpu_csr_towb.core_memsize_i <= std_logic_vector(to_unsigned(g_config.cpu_memsizes(cpu_index), 32));

  cpu_csr_towb.dbg_mbx0_i <= mbx_data_towb(0);
  cpu_csr_towb.dbg_mbx1_i <= mbx_data_towb(1);
  mbx_data_fromwb <= (0 => cpu_csr_fromwb.dbg_mbx0_o,
                      1 => cpu_csr_fromwb.dbg_mbx1_o,
                      others => x"0000_0000");
  mbx_wr <= (0 => cpu_csr_fromwb.dbg_mbx0_load_o,
             1 => cpu_csr_fromwb.dbg_mbx1_load_o,
             others => '0');
  dbg_insn <= (0 => cpu_csr_fromwb.dbg_cpu0_insn_o,
               1 => cpu_csr_fromwb.dbg_cpu1_insn_o,
               others => x"0000_0000");
  dbg_insn_wr <= (0 => cpu_csr_fromwb.dbg_cpu0_insn_wr_o,
                  1 => cpu_csr_fromwb.dbg_cpu1_insn_wr_o,
                  others => '0');

  gen_cpus : for i in 0 to g_config.cpu_count-1 generate

    U_CPU_Block : entity work.mt_cpu_cb
      generic map (
        g_cpu_id            => i,
        g_iram_size         => g_config.cpu_memsizes(i),
        g_with_white_rabbit => g_with_white_rabbit,
        g_system_clock_freq => g_system_clock_freq)
      port map (
        clk_sys_i   => clk_i,
        rst_n_i     => rst_n_i,
        clk_ref_i   => clk_ref_i,
        rst_n_ref_i => rst_n_ref,
        tm_i        => tm_i,
        sh_master_i => si_slave_out(c_si_slave_cpu0 + i),
        sh_master_o => si_slave_in(c_si_slave_cpu0 + i),
        dp_master_i => dp_master_i(i),
        dp_master_o => dp_master_o(i),
        cpu_csr_i   => per_cpu_csr_fromwb(i),
        cpu_csr_o   => per_cpu_csr_towb(i),
        rmq_ready_i => rmq_status,
        hmq_ready_i => hmq_status,
        gpio_o      => cpu_gpio_out(i),
        gpio_i      => gpio_i,
        pc_o        => cpu_pcs(i),
        pc_valid_o  => cpu_pcs_valid(i),
        dbg_drdy_o  => cpu_dbg_drdy(i),
        dbg_dack_i  => cpu_dbg_dack(i),
        dbg_data_o  => cpu_dbg_msg_data(i));

    cpu_csr_towb.dbg_cpu_status_i(i) <= per_cpu_csr_towb(i).dbg_cpu_status_i;
    cpu_csr_towb.dbg_cpu_insn_ready_i(i) <= per_cpu_csr_towb(i).dbg_insn_ready_i;
    cpu_csr_towb.dbg_poll_ready_i(i) <= cpu_dbg_drdy(i);

    mbx_data_towb(i) <= per_cpu_csr_towb(i).dbg_mbx_i;

    per_cpu_csr_fromwb(i) <=
      (reset_o           => cpu_csr_fromwb.reset_o(i),
       enable_o          => cpu_csr_fromwb.enable_o(i),
       uaddr_addr_o      => cpu_csr_fromwb.uaddr_addr_o,
       core_sel_o        => f_to_std_logic(cpu_index = i),
       udata_o           => cpu_csr_fromwb.udata_o,
       udata_load_o      => cpu_csr_fromwb.udata_load_o,
       dbg_cpu_force_o   => cpu_csr_fromwb.dbg_cpu_force_o(i),
       dbg_insn_o        => dbg_insn(i),
       dbg_insn_wr_o     => dbg_insn_wr(i),
       dbg_mbx_o         => mbx_data_fromwb(i),
       dbg_mbx_wr_o      => mbx_wr(i));
  end generate gen_cpus;

  gen_no_cpus : for i in g_config.cpu_count to t_max_cpu_rng'high generate
    cpu_csr_towb.dbg_cpu_status_i(i) <= '0';
    cpu_csr_towb.dbg_cpu_insn_ready_i(i) <= '0';
    cpu_csr_towb.dbg_poll_ready_i(i) <= '0';
  end generate gen_no_cpus;

  cpu_csr_towb.udata_i <= per_cpu_csr_towb(cpu_index).udata_i;

  U_Host_MQ : mt_mqueue_host
    generic map (
      g_config => g_config.hmq_config)
    port map (
      clk_i        => clk_i,
      rst_n_i      => rst_n_i,
      si_slave_i   => si_master_out(c_si_master_hmq),
      si_slave_o   => si_master_in(c_si_master_hmq),
      host_slave_i => hac_master_out(c_hac_master_hmq),
      host_slave_o => hac_master_in(c_hac_master_hmq),
      host_irq_o   => host_irq_o,
      hmq_status_o => hmq_status);

  gen_with_rmq : if g_with_rmq generate

    U_Remote_MQ : mt_mqueue_remote
      generic map (
        g_config        => g_config.rmq_config,
        g_use_wr_fabric => g_use_wr_fabric
        )
      port map (
        clk_i        => clk_i,
        rst_n_i      => rst_n_i,
        rmq_swrst_o  => rmq_swrst_o,
        si_slave_i   => si_master_out(c_si_master_rmq),
        si_slave_o   => si_master_in(c_si_master_rmq),
        src_o        => rmq_src_o,
        src_i        => rmq_src_i,
        snk_i        => rmq_snk_i,
        snk_o        => rmq_snk_o,
        wr_src_o     => wr_src_o,
        wr_src_i     => wr_src_i,
        wr_snk_o     => wr_snk_o,
        wr_snk_i     => wr_snk_i,
        rmq_status_o => rmq_status,
        debug_o      => open);


  end generate gen_with_rmq;

  gen_without_rmq : if not g_with_rmq generate

    rmq_status                    <= (others => '0');
    si_master_in(c_si_master_rmq) <= cc_dummy_slave_out;

  end generate gen_without_rmq;




  U_Shared_Mem : mt_shared_mem
    generic map (
      g_size => g_config.shared_mem_size / 4)
    port map (
      clk_i   => clk_i,
      rst_n_i => rst_n_i,
      slave_i => si_master_out(c_si_master_smem),
      slave_o => si_master_in(c_si_master_smem));

  -- Configuration ROM with two simultaneous read-only ports
  -- for parallel access from HAC and SI crossbars
  U_Config_Rom: entity work.mt_config_rom
    generic map (
      g_config            => g_config,
      g_with_rmq          => g_with_rmq,
      g_system_clock_freq => g_system_clock_freq,
      g_with_white_rabbit => g_with_white_rabbit)
    port map (
      clk_i    => clk_i,
      rst_n_i  => rst_n_i,
      slave1_i => hac_master_out(c_hac_master_crom),
      slave1_o => hac_master_in(c_hac_master_crom),
      slave2_i => si_master_out(c_si_master_crom),
      slave2_o => si_master_in(c_si_master_crom));

  p_mux_debug_data : process(cpu_dbg_msg_data, cpu_csr_fromwb, dbg_msg_data_read_ack)
  begin

    cpu_csr_towb.dbg_msg_data_i <= (others => 'X');

    for i in 0 to g_config.cpu_count-1 loop
      if unsigned(cpu_csr_fromwb.core_sel_o) = to_unsigned(i, 4) then
        cpu_csr_towb.dbg_msg_data_i <= cpu_dbg_msg_data(i);
        cpu_dbg_dack(i)             <= dbg_msg_data_read_ack;
      else
        cpu_dbg_dack(i) <= '0';
      end if;
    end loop;
  end process;

  p_debug_irq : process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        debug_msg_irq_o <= '0';
      else
        debug_msg_irq_o <= '0';

        for i in 0 to g_config.cpu_count-1 loop
          if (cpu_dbg_drdy(i) and cpu_csr_fromwb.dbg_imsk_enable_o(i)) = '1'
          then
            debug_msg_irq_o <= '1';
          end if;
        end loop;
      end if;
    end if;
  end process p_debug_irq;

  gpio_o <= f_reduce_or(cpu_gpio_out);

end rtl;
