/*
 * Copyright (C) 2015-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <errno.h>
#include <mockturtle-framework.h>

uint32_t msg_seq = 0;
struct rt_application *_app;


/**
 * Some Mock Turtle version support only 32bit addressing.
 * This memcpy implementation perform a 32bit aligned copy
 * @param[out] dest destination buffer
 * @param[in] src source buffer
 * @param[in] n number of bytes
 */
void rt_memcpy(uint32_t *dest, uint32_t *src, size_t n)
{
	int i;

	for (i = 0; i < n / 4; ++i)
		dest[i] = src[i];
}


/**
 * It get the current time from the internal WRNC timer
 * @param[out] seconds
 * @param[out] cycles
 */
void rt_get_time(uint32_t *seconds, uint32_t *cycles)
{
	*seconds = lr_readl(MT_CPU_LR_REG_TAI_SEC);
	*cycles = lr_readl(MT_CPU_LR_REG_TAI_CYCLES);
}


/**
 * It initilizes the declared MQs
 * @param[in] app application declaration
 * @return 0 on success. -1 on error
 */
static inline int trtl_rt_init_mq(void)
{
	int i;

	for (i = 0; i < _app->n_mq; ++i) {
		if (_app->mq[i].flags & TRTL_RT_MQ_FLAGS_OUTPUT) {
			_app->mq[i].buf = MQ_BUFFER_OUT(_app->mq[i].type,
							_app->mq[i].index);
			mq_purge(_app->mq[i].type, _app->mq[i].index);
			if (_app->mq[i].flags & TRTL_RT_MQ_FLAGS_CLAIM) {
				mq_claim(_app->mq[i].type, _app->mq[i].index);
			}
		} else {
			_app->mq[i].buf = MQ_BUFFER_IN(_app->mq[i].type,
						       _app->mq[i].index);
		}
	}

	return 0;
}


/**
 * The function initializes the library and it does some compatibility check.
 *
 * Initialization
 * - purge all the output host-message-queue
 *
 * compatibility checks:
 * - check bitstream FPGA ID if compatibile with the application. The value in
 *   app->version.fpga_id will be updated with the FPGA ID from the bitstream
 *
 * @param[in] app application declaration
 * @return 0 on success. -1 on error
 */
int rt_init(struct rt_application *app)
{
	int i, err;
	uint32_t *fpga_id = &app->version.fpga_id; /* in order to overwrite the const */

	_app = app;

	/* Validate actions */
	for (i = 0; i < _app->n_actions; ++i) {
		if (i > __TRTL_MSG_ID_MAX_USER) {
			pp_printf("Too many actions defined. Maximum '%d' user defined actions\n\r",
				  __TRTL_MSG_ID_MAX_USER);
			return -1;
		}
		if (_app->actions[i] == NULL) {
			pp_printf("Action %d not defined\n\r", i);
			return -1;
		}
	}

	pp_printf("Running application '%s'\n\r", _app->name);

	/* *fpga_id = 0x1234; /\* TODO read it from the FPGA ID register *\/ */
	for (i = 0; i < app->fpga_id_compat_n; i++) {
		if (app->fpga_id_compat[i] == *fpga_id)
			break;
	}
	if (app->fpga_id_compat_n && app->fpga_id_compat_n == i) {
		pp_printf("  FPGA '0x%"PRIx32"' not compatible with RT app: 0x%"PRIx32" 0x%"PRIx32" (git %"PRIx32")\n\r",
			  _app->version.fpga_id,
			  _app->version.rt_id,
			  _app->version.rt_version,
			  _app->version.git_version);
		return -1;
	}
	pp_printf("  application id     \t'0x%"PRIx32"'\n\r", _app->version.rt_id);
	pp_printf("  application version\t'%"PRId32".%"PRId32"'\n\r",
		  RT_VERSION_MAJ(_app->version.rt_version),
		  RT_VERSION_MIN(_app->version.rt_version));
	pp_printf("  source code id     \t'0x%"PRIx32"'\n\r", _app->version.git_version);

	err = trtl_rt_init_mq();
	if (err)
		return err;

#ifdef LIBRT_DEBUG
	pp_printf("Exported Variables\n\r");
	for (i = 0; i < _app->n_variables; ++i)
		pp_printf("[%d] addr 0x%x | mask 0x%x | off %d\n\r", i,
			  _app->variables[i].addr,
			  _app->variables[i].mask,
			  _app->variables[i].offset);
	pp_printf("Exported Buffers\n\r");
	for (i = 0; i < _app->n_buffers; ++i)
		pp_printf("[%d] addr %p | len %d\n\r", i,
			  _app->buffers[i].buf,
			  _app->buffers[i].len);
#endif

	return 0;
}
