/**
 * @defgroup library-serial Serial Console
 * @{
 * @ingroup library
 * @brief Serial Console functions
 * @copyright (C) 2013-2016 CERN (www.cern.ch)
 * @author Alessandro Rubini <rubini@gnudd.com>
 * @author Federico Vaga <federico.vaga@cern.ch>
 *
 * Released according to the GNU GPL, version 2 or any later version.
 */

#ifndef __TRTL_RT_PP_PRINTF_H
#define __TRTL_RT_PP_PRINTF_H

#include <stdarg.h>

/**
 * Print buffer size. This means that your printf string
 * should not exceed this value. If the string exceed this value
 * then it will be trouncated.
 */
#define CONFIG_PRINT_BUFSIZE 128


/**
 * It prints a string on the serial interface.
 * Internally, it uses the puts() function.
 * @param[in] fmt string format
 * @param[in] ... argument according to the string format
 * @return number of printed characters
 */
#ifdef LIB_PRINTF
extern int pp_printf(const char *fmt, ...)
        __attribute__((format(printf,1,2)));
#else
static inline int pp_printf(const char *fmt, ...)
{
        return 0;
}
#endif


/**
 * It creates a new string according to the given format
 * @param[out] s output string
 * @param[in] fmt string format
 * @param[in] ... argument according to the string format
 * @return number of printed characters
 */
#ifdef LIB_PRINTF
extern int pp_sprintf(char *s, const char *fmt, ...)
        __attribute__((format(printf,2,3)));
#else
static inline int pp_sprintf(char *s, const char *fmt, ...)
{
        return 0;
}
#endif


/**
 * It prints a string on the serial interface.
 * Internally, it uses the puts() function.
 * @param[in] fmt string format
 * @param[in] args list of arguments according to the string format
 * @return number of printed characters
 */
#ifdef LIB_PRINTF
extern int pp_vprintf(const char *fmt, va_list args);
#else
static inline int pp_vprintf(const char *fmt, va_list args)
{
        return 0;
}
#endif


/**
 * It creates a new string according to the given format
 * @param[out] buf output string
 * @param[in] fmt string format
 * @param[in] args list of arguments according to the string format
 * @return number of printed characters
 */
#ifdef LIB_PRINTF
extern int pp_vsprintf(char *buf, const char *, va_list args)
        __attribute__ ((format (printf, 2, 0)));
#else
static inline int pp_vsprintf(char *buf, const char *fmt, va_list args)
{
        return 0;
}
#endif


/**
 * It prints a string on the serial interface only when the support
 * for error messages is enable.
 *
 * Standard Makefile -> MOCKTURTLE_LIBRARY_PRINT_DEBUG_ENABLE
 *
 * C Preprocessor ->LIB_PRINT_DEBUG
 *
 * Internally, it uses the puts() function.
 * @param[in] fmt string format
 * @param[in] ... argument according to the string format
 * @return number of printed characters
 */
#if defined(LIB_PRINTF) && defined(LIB_PRINT_DEBUG)
extern int pr_debug(const char *fmt, ...)
        __attribute__((format(printf,1,2)));
#else
static inline int pr_debug(const char *fmt, ...)
{
        return 0;
}
#endif

/**
 * It prints a string on the serial interface only when the support
 * for error messages is enable.
 *
 * Standard Makefile -> MOCKTURTLE_LIBRARY_PRINT_ERROR_ENABLE
 *
 * C Preprocessor -> LIB_PRINT_ERROR
 *
 * Internally, it uses the puts() function.
 * @param[in] fmt string format
 * @param[in] ... argument according to the string format
 * @return number of printed characters
 */
#if defined(LIB_PRINTF) && defined(LIB_PRINT_ERROR)
extern int pr_error(const char *fmt, ...)
        __attribute__((format(printf,1,2)));
#else
static inline int pr_error(const char *fmt, ...)
{
        return 0;
}
#endif

extern int puts(const char *p);

/**@}*/

#endif
